#include <math.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include "XPLMDisplay.h"
#include "XPLMGraphics.h"
#include "XPLMDataAccess.h"
#include "XPLMUtilities.h"
#include "XPLMProcessing.h"
#include "XPLMMenus.h"
#include "XPLMNavigation.h"

#ifdef __APPLE__
#include <OpenGL/gl.h>
#include <Carbon/Carbon.h>
#else
#include <stdlib.h>
#include <GL/gl.h>
#endif


static float gameLoopCallback(float inElapsedSinceLastCall,
                float inElapsedTimeSinceLastFlightLoop, int inCounter,    
                void *inRefcon);
                

#define STATE_LAND 1
#define STATE_AIR 2

#define WINDOW_WIDTH 130
#define WINDOW_HEIGHT 95

#define MAX_TOKEN_BUF 100

#define MAX_GVALUES 10

#define MSG_COUNT 5
#define MSG_LENGTH 100

struct LandingInfo {
	float landingG;
	float landingVSpeed;
};

struct LandingParams {
	float lastVSpeed;
	float lastVG;
	int curGIdx;
	float lastGVals[120];
	int lastState;
};

struct PluginParams {
	XPLMMenuID xgsMenu; // = NULL;
	int enableLogItem;
};

struct DataRefs {
	XPLMDataRef vertSpeedRef;
	XPLMDataRef vertGRef;
	XPLMDataRef gearKoofRef;
	XPLMDataRef flightTimeRef;
	XPLMDataRef descrRef;

	XPLMDataRef longRef;
	XPLMDataRef latRef;
	XPLMDataRef craftNumRef;
};

struct Window {
	int winPosX /*= 20 */;
	int winPosY /*= 600 */;
	int lastMouseX;
	int lastMouseY;
	XPLMWindowID gWindow;// = NULL;
	float remainingShowTime;// = 0.0f;
	int windowCloseRequest; // = 0;
};

struct LogData {
	char logAirportId[50];
	char logAirportName[300];
	char logAircraftNum[50];
	time_t landingTime;

};

struct LogParams {
	int logThisLanding;
	int logEnabled;
};

/* not a good idea to do so. need to be cleaned up */
static struct LandingInfo lndInfo;
static struct LandingParams lndParams;
static struct PluginParams pluginParams;
static struct DataRefs dataRefs;
static struct Window wnd;
static struct LogData logData;
static struct LogParams logParams;

/* not a good idea to do so. need to be cleaned up */
static char logMsg[MSG_COUNT][MSG_LENGTH];

#ifdef __APPLE__
int MacToUnixPath(const char * inPath, char * outPath, int outPathMaxLen)
{
    CFStringRef inStr = CFStringCreateWithCString(kCFAllocatorDefault, inPath, kCFStringEncodingMacRoman);
    if (inStr == NULL) return -1;
    CFURLRef url = CFURLCreateWithFileSystemPath(kCFAllocatorDefault, inStr, kCFURLHFSPathStyle,0);
    CFStringRef outStr = CFURLCopyFileSystemPath(url, kCFURLPOSIXPathStyle);
    if (!CFStringGetCString(outStr, outPath, outPathMaxLen, kCFURLPOSIXPathStyle)) return -1;
    CFRelease(outStr);
    CFRelease(url);
    CFRelease(inStr);
    return 0;
}
#endif


static FILE* getConfigFile(char *mode)
{
    char path[512];
    
    XPLMGetPrefsPath(path);
    XPLMExtractFileAndPath(path);
    strcat(path, XPLMGetDirectorySeparator());
    strcat(path, "xgs.prf");

#ifdef __APPLE__
    char unixPath[512];
    MacToUnixPath(path, unixPath, 512);
    
    return fopen(unixPath, mode);
#else
    return fopen(path, mode);
#endif
}


static int skipSpaces(FILE *f)
{
    int ch;
    
    do {
        ch = fgetc(f);
        if (EOF == ch)
            return -1;
    } while (isspace(ch));
    ungetc(ch, f);
    
    return 0;
}


static int getNextToken(FILE *f, char *buf)
{
    int pos = 0;
    int ch;
    
    if (skipSpaces(f))
        return -1;
    
    ch = fgetc(f);
    while ((! isspace(ch)) && (EOF != ch) && (MAX_TOKEN_BUF - 1 > pos)) {
        buf[pos++] = ch;
        ch = fgetc(f);
    }
    if (MAX_TOKEN_BUF - 1 == pos)
        return -1;
    
    buf[pos] = 0;
    return 0;
}


static int tokenToInt(char *token, int *val)
{
    char *endptr;
    int v = strtol(token, &endptr, 10);
    if (endptr && (! *endptr)) {
        *val = v;
        return 0;
    } else
        return -1;
}

static void saveConfig()
{
    FILE *f;
    
    f = getConfigFile("w");
    if (! f)
        return;
    
    fprintf(f, "%i %i %i", wnd.winPosX, wnd.winPosY, logParams.logEnabled);
    
    fclose(f);
}


#define ABORT_CONF { fclose(f); return; }

static void loadConfig()
{
    FILE *f;
    char s[MAX_TOKEN_BUF];
    
    f = getConfigFile("r");
    if (! f)
        return;
    
    if (getNextToken(f, s)) ABORT_CONF
    tokenToInt(s, &wnd.winPosX);
    if (getNextToken(f, s)) ABORT_CONF
    tokenToInt(s, &wnd.winPosY);
    if (getNextToken(f, s)) ABORT_CONF
    tokenToInt(s, &logParams.logEnabled);
    
    fclose(f);
}

#undef ABORT_CONF


static void updateLogItemState()
{
    XPLMCheckMenuItem(pluginParams.xgsMenu, pluginParams.enableLogItem, 
        logParams.logEnabled ? xplm_Menu_Checked : xplm_Menu_Unchecked);
}


static void xgsMenuCallback(void *menuRef, void *param)
{
    logParams.logEnabled = ! logParams.logEnabled;
    updateLogItemState();
}


static void initDataRefs(struct DataRefs *pdr)
{
	pdr->vertGRef = XPLMFindDataRef("sim/flightmodel/forces/g_nrml");
    pdr->vertSpeedRef = XPLMFindDataRef("sim/flightmodel/position/vh_ind");
    pdr->gearKoofRef = XPLMFindDataRef("sim/flightmodel/forces/faxil_gear");
    pdr->flightTimeRef = XPLMFindDataRef("sim/time/total_flight_time_sec");
    pdr->descrRef = XPLMFindDataRef("sim/aircraft/view/acf_tailnum");
    pdr->latRef = XPLMFindDataRef("sim/flightmodel/position/latitude");
    pdr->longRef = XPLMFindDataRef("sim/flightmodel/position/longitude");
    pdr->craftNumRef = XPLMFindDataRef("sim/aircraft/view/acf_tailnum");
}

static void initWindow(struct Window *pwnd)
{
	memset(pwnd, 0, sizeof(struct Window));
	pwnd->winPosX = 20;
	pwnd->winPosY = 600;

}

PLUGIN_API int XPluginStart(char *outName, char *outSig, char *outDesc)
{
    XPLMMenuID pluginsMenu;
    int subMenuItem;
    
    loadConfig();
    
    strcpy(outName, "Landing Speed and G-Load");
    strcpy(outSig, "inv.landspeedG");
    strcpy(outDesc, "A plugin that shows vertical landing speed and G forces.");

	initDataRefs(&dataRefs);
	initWindow(&wnd);

	memset(&lndInfo, 0, sizeof(struct LandingInfo));
	memset(&lndParams, 0, sizeof(struct LandingParams));
	memset(&pluginParams, 0, sizeof(struct PluginParams));
	memset(&logData, 0, sizeof(struct LogData));
	memset(&logParams, 0, sizeof(struct LogParams));

	memset(logMsg, MSG_COUNT*MSG_LENGTH, 0);
    
    XPLMRegisterFlightLoopCallback(gameLoopCallback, 0.05f, NULL);
    
    pluginsMenu = XPLMFindPluginsMenu();
    subMenuItem = XPLMAppendMenuItem(pluginsMenu, "Landing Speed G", NULL, 1);

    pluginParams.xgsMenu = XPLMCreateMenu("Landing Speed G", pluginsMenu, subMenuItem, 
                xgsMenuCallback, NULL);     
    pluginParams.enableLogItem = XPLMAppendMenuItem(pluginParams.xgsMenu, "Enable Log", NULL, 1);

    updateLogItemState();
    
    return 1;
}


static void trim(char *str)
{
    int len = strlen(str);
    len--;
    while (0 < len) {
        if (('\r' == str[len]) || ('\n' == str[len])) {
            str[len] = 0;
            len--;
        } else
            return;
    }
}


static void writeLandingToLog()
{
    FILE *f;
    char buf[512];
    
    logParams.logThisLanding = 0;
    XPLMGetSystemPath(buf);
    strcat(buf, "landing.log");

#ifdef __APPLE__
    char unixPath[512];
    MacToUnixPath(buf, unixPath, 512);

    f = fopen(unixPath, "a");
#else
    f = fopen(buf, "a");
#endif
    if (! f) return;

    strcpy(buf, ctime(&logData.landingTime));
    trim(buf);
    fprintf(f, "%s %s %s '%s' %.3f m/s %i fpm %s\n", buf, logData.logAircraftNum,
                logData.logAirportId, logData.logAirportName, fabs(lndInfo.landingVSpeed),
                abs(rint(lndInfo.landingVSpeed * 60.0f * 3.2808f)), logMsg[0]);

    fclose(f);
}


static void closeEventWindow()
{
    if (logParams.logThisLanding)
        writeLandingToLog();
    
    if (wnd.gWindow) {
        XPLMDestroyWindow(wnd.gWindow);
        wnd.gWindow = NULL;
    }
    
    lndInfo.landingVSpeed = 0.0f;
    lndInfo.landingG = 0.0f;
    lndParams.curGIdx = 0;
    int i = 0;
    for(i = 0; i < MSG_COUNT; ++i)
		logMsg[i][0] = 0;
    wnd.remainingShowTime = 0.0f;
}

PLUGIN_API void	XPluginStop(void)
{
    closeEventWindow();
    saveConfig();
}


PLUGIN_API void XPluginDisable(void)
{
}


PLUGIN_API int XPluginEnable(void)
{
	return 1;
}


PLUGIN_API void XPluginReceiveMessage(XPLMPluginID inFromWho,
                long inMessage,	void *inParam)
{
}

void MyDrawWindowCallback(XPLMWindowID inWindowID, void *inRefcon)
{
    if (0.0f < wnd.remainingShowTime) {
        int left, top, right, bottom;
        float color[] = { 1.0, 1.0, 1.0 }; 	/* RGB White */
            
        XPLMGetWindowGeometry(inWindowID, &left, &top, &right, &bottom);
        XPLMDrawTranslucentDarkBox(left, top, right, bottom);
        XPLMDrawString(color, left + 5, top - 20, logMsg[0], NULL, xplmFont_Basic);
        XPLMDrawString(color, left + 5, top - 35, logMsg[1], NULL, xplmFont_Basic);
        XPLMDrawString(color, left + 5, top - 50, logMsg[2], NULL, xplmFont_Basic);
        XPLMDrawString(color, left + 5, top - 65, logMsg[3], NULL, xplmFont_Basic);
        
        glDisable(GL_TEXTURE_2D);
        glColor3f(0.7, 0.7, 0.7);
        glBegin(GL_LINES);
          glVertex2i(right - 1, top - 1);
          glVertex2i(right - 7, top - 7);
          glVertex2i(right - 7, top - 1);
          glVertex2i(right - 1, top - 7);
        glEnd();

        glColor3f(1.0, 0.0, 0.0);
        int i = 0;
        glBegin(GL_LINES);
          for(i = 0; i < MAX_GVALUES-1; ++i) {
			glVertex2i(left + 15 + 2*i - 1, top - 93 + lndParams.lastGVals[i]*5);
			glVertex2i(left + 15 + 2*(i + 1) - 1, top - 93 + lndParams.lastGVals[i+1]*5);
		  }
        glEnd();
        
        glEnable(GL_TEXTURE_2D);
    }
}                                   

static int mouseCallback(XPLMWindowID inWindowID, int x, int y,    
                   XPLMMouseStatus inMouse, void *inRefcon)
{
    if (wnd.windowCloseRequest)
        return 1;
    
    switch (inMouse) {
        case xplm_MouseDown:
            if ((x >= wnd.winPosX + WINDOW_WIDTH - 8) && (x <= wnd.winPosX + WINDOW_WIDTH) && 
                        (y <= wnd.winPosY) && (y >= wnd.winPosY - 8))
                wnd.windowCloseRequest = 1;
            else {
                wnd.lastMouseX = x;
                wnd.lastMouseY = y;
            }
            break;
            
        case xplm_MouseDrag:
            wnd.winPosX += x - wnd.lastMouseX;
            wnd.winPosY += y - wnd.lastMouseY;
            XPLMSetWindowGeometry(wnd.gWindow, wnd.winPosX, wnd.winPosY, 
                    wnd.winPosX + WINDOW_WIDTH, wnd.winPosY - WINDOW_HEIGHT);
            wnd.lastMouseX = x;
            wnd.lastMouseY = y;
            break;
            
        case xplm_MouseUp:
            break;
    }

    return 1;
}

static void keyboardCallback(XPLMWindowID inWindowID, char inKey, XPLMKeyFlags inFlags,    
                   char inVirtualKey, void *inRefcon, int losingFocus)
{
}

static int getCurrentState()
{
    return 0.0f != XPLMGetDataf(dataRefs.gearKoofRef) ? STATE_LAND : STATE_AIR;
}


static void gradeMessage(float gn, char* grade)
{
	if (gn < 1.2)
        sprintf(grade, "excelent landing");
    else if (gn < 1.5)
        sprintf(grade, "good landing");
    else if (gn < 1.8)
        sprintf(grade, "acceptable landing");
    else if (gn < 2.5)
        sprintf(grade, "hard landing");
    else if (gn < 6.0)
        sprintf(grade, "you are fired!!!");
    else if (gn < 20.0)
        sprintf(grade, "anybody survived?");
	else
        sprintf(grade, "R.I.P.");

}


static void prepareToLog()
{
    int num;
    float lat = XPLMGetDataf(dataRefs.latRef);
    float lon = XPLMGetDataf(dataRefs.longRef);
    XPLMNavRef ref = XPLMFindNavAid(NULL, NULL, &lat, &lon, NULL, xplm_Nav_Airport);
    
    if (XPLM_NAV_NOT_FOUND != ref)
        XPLMGetNavAidInfo(ref, NULL, &lat, &lon, NULL, NULL, NULL, logData.logAirportId, 
                logData.logAirportName, NULL);
    else {
        logData.logAirportId[0] = 0;
        logData.logAirportName[0] = 0;
    }
    logData.landingTime = time(NULL);
    num = XPLMGetDatab(dataRefs.craftNumRef, logData.logAircraftNum, 0, 49);
    logData.logAircraftNum[num] = 0;
    logParams.logThisLanding = 1;
}


/*
 * for now trivial O(n) selection algorithm to select a maximum G
 */
static float getMaxG(float *g, unsigned size)
{
	float resG = 0;
	int i = 0;
	for(i = 0; i < size; ++i)
	{
		if(g[i] > resG)
			resG = g[i];
	}
	return resG;
}


static void createEventWindow()
{
	lndInfo.landingG = getMaxG(lndParams.lastGVals, MAX_GVALUES);
    gradeMessage(lndInfo.landingG, logMsg[0]);
    sprintf(logMsg[1], "Vy: %i fpm", abs(rint(lndInfo.landingVSpeed * 60.0f * 3.2808f)));
    sprintf(logMsg[2], "Vy: %.3f m/s", fabs(lndInfo.landingVSpeed));
    sprintf(logMsg[3], "Gy max: %.2f", lndInfo.landingG);

    if (! wnd.gWindow)
        wnd.gWindow = XPLMCreateWindow(wnd.winPosX, wnd.winPosY, 
                    wnd.winPosX + WINDOW_WIDTH, wnd.winPosY - WINDOW_HEIGHT, 
                    1, MyDrawWindowCallback, keyboardCallback, 
                    mouseCallback, NULL);
    if (logParams.logEnabled && (! logParams.logThisLanding))
        prepareToLog();
}

static float gameLoopCallback(float inElapsedSinceLastCall,
                float inElapsedTimeSinceLastFlightLoop, int inCounter,    
                void *inRefcon)
{
    int state = getCurrentState();
    float timeFromStart = XPLMGetDataf(dataRefs.flightTimeRef);

    if (3.0 < timeFromStart) {
        if (wnd.windowCloseRequest) {
            wnd.windowCloseRequest = 0;
            closeEventWindow();
        } else
            if (0.0f < wnd.remainingShowTime) {
                wnd.remainingShowTime -= inElapsedSinceLastCall;
                if (0.0f >= wnd.remainingShowTime)
                    closeEventWindow();
            }
    
        if ((STATE_AIR == lndParams.lastState) && (STATE_LAND == state)) {
            if (lndParams.lastVSpeed < lndInfo.landingVSpeed) {
                lndInfo.landingVSpeed = lndParams.lastVSpeed;
            }
            if(lndParams.curGIdx < MAX_GVALUES)
				lndParams.lastGVals[lndParams.curGIdx++] = lndParams.lastVG;
            wnd.remainingShowTime = 60.0f;
        }
        else if(STATE_LAND == state && wnd.remainingShowTime > 0.0) {
			/* collect G values after a touchdown for analysis. than create window and analize */
            if(lndParams.curGIdx < MAX_GVALUES)
				lndParams.lastGVals[lndParams.curGIdx++] = lndParams.lastVG;
			else
                createEventWindow();
			
		}
    }
    
    lndParams.lastVSpeed = XPLMGetDataf(dataRefs.vertSpeedRef);
    lndParams.lastVG = XPLMGetDataf(dataRefs.vertGRef);
    lndParams.lastState = state;
    return 0.05f;
}

